#!/bin/bash
#
# Update the internal_users via 'securityadmin.sh' to set their passwords.
# This should only be called on a new cluster.
#
# Usage: docker compose run os-node-1 ./init-security.sh
#

set -euo pipefail

# Define common paths in the opensearch docker container.
security_config_path="./config/opensearch-security"
security_tools_path="./plugins/opensearch-security/tools"

# Get the passwords from the env vars.
admin_pw="${OPENSEARCH_ADMIN_PASSWORD:?admin password required}"
kibanaserver_pw="${OPENSEARCH_KIBANASERVER_PASSWORD:?kibana server password required}"
logginguser_pw="${OPENSEARCH_LOGGINGUSER_PASSWORD:?logging user password required}"

# Hash the passwords using the security plugin's hash script.
admin_hash=$(${security_tools_path}/hash.sh -p "${admin_pw}" | awk '{print $NF}')
dashboards_hash=$(${security_tools_path}/hash.sh -p "${kibanaserver_pw}" | awk '{print $NF}')
logging_hash=$(${security_tools_path}/hash.sh -p "${logginguser_pw}" | awk '{print $NF}')

internal_users_file="${security_config_path}/internal_users.yml"
internal_users_template="${security_config_path}/internal_users.template.yml"

# Create internal_users.yml from the template.
# This is important to prevent docker mount issues.
cp "${internal_users_template}" "${internal_users_file}"
echo "[INFO] Updating password hashes."

# Create a temporary file before modifying.
cp "${internal_users_file}" "${internal_users_file}.tmp"

# Add an autogenerated comment and remove the original 1st and 2nd lines.
{
    echo "# ------------------------------------------------- #"
    echo "# Auto-generated from 'internal_users.template.yml' #"
    echo "# ------------------------------------------------- #"
    sed '1d;2d' "${internal_users_file}"
} > "${internal_users_file}.tmp"

# Update the password hashes while preserving all other content.
sed -E \
    -e "s|^( *hash: )[^\s]+(  # ADMIN_PW)|\1${admin_hash}\2|" \
    -e "s|^( *hash: )[^\s]+(  # DASHBOARDS_PW)|\1${dashboards_hash}\2|" \
    -e "s|^( *hash: )[^\s]+(  # LOGGING_PW)|\1${logging_hash}\2|" \
    "${internal_users_file}.tmp" > "${internal_users_file}"

# Manual cleanup as we persist configuration (see 'docker-compose.base.yml').
rm "${internal_users_file}.tmp"

# Start OpenSearch (securityadmin.sh needs the security index).
opensearch -Eplugins.security.allow_default_init_securityindex=true -Ediscovery.type=single-node &
os_pid=$!
retries=10

# Wait until OpenSearch is ready to accept connections.
while [[ "${retries}" -gt 0 ]] ; do
    echo "[INFO] Waiting for opensearch to be up. $retries"
    sleep 5s
    if curl -ksu "admin:admin" "https://localhost:9200"; then
      break
    fi
    retries=$(( ${retries} - 1 ))
done

if [[ "${retries}" -eq 0 ]]; then
  echo &>2 "[ERROR] Exceeded maximum number of retries!"
  exit 1
fi

# Wait for opensearch to be healthy.
curl -ku "admin:admin" "https://localhost:9200/_cluster/health?wait_for_status=yellow"

# The securityadmin script needs admin certificates or a keystore to operate.
"${security_tools_path}/securityadmin.sh" \
  -rev -icl -nhnv \
  -cacert ./config/root-ca.pem \
  -cert ./config/admin-crt.pem \
  -key ./config/admin-key.pem \
  -cd "${security_config_path}"

# Shut down opensearch again.
sleep 5s
kill ${os_pid}
