#!/bin/bash
#
# Script for generating SSL keys and certificates.
#

set -euo pipefail

# The guard below ensures, in most cases, that the script is executed from the correct directory (project directory).
# So it should be safe to assume that we can extract the path from the parent of the 'scripts' directory.
project_dir="$(dirname "$(dirname "$(realpath "$0")")")"

# ---------------- #
# Execution checks #
# ---------------- #
if [[ -d "ssl" ]]; then
    cd "ssl"
else
    echo >&2 "[ERROR] This script needs to be executed from the project directory!"
    exit 1
fi

# --------------------- #
# Environment variables #
# --------------------- #
env_file="${project_dir}/.env"

# Check if the .env file exists and source it
if [[ -f "${env_file}" ]]; then
    echo "[INFO] Detected .env file, sourcing it."
    source "${env_file}"
fi

# Set hostnames for Dashboards and Opensearch nodes.
# They are determining whether the setup is in a local or live environment.
dashboards_hostname="${DASHBOARDS_HOSTNAME:-localhost}"
opensearch_hostname="${OPENSEARCH_HOSTNAME:-localhost}"

certbot_email="${CERTBOT_EMAIL:-}"
certbot_rsa_key_size="${CERTBOT_RSA_KEY_SIZE:-4096}"
certbot_staging="${CERTBOT_STAGING:-}"

# --------------------- #
# Utilities definitions #
# --------------------- #
# Common certificate details
organization="TU Wien"
org_unit="Center for Research Data Management"
country="AT"
state="Vienna"
locality="Vienna"

function generate_certificate {
    # Generate keypair
    local cert_name="${1}"
    local common_name="${2}"

    # Generate the ssl files
    subject="/C=${country}/ST=${state}/L=${locality}/O=${organization}/OU=${org_unit}/CN=${common_name}"
    openssl genrsa -out "${cert_name}-key-temp.pem" 2048
    openssl pkcs8 -inform PEM -outform PEM -in "${cert_name}-key-temp.pem" -topk8 -nocrypt -v1 PBE-SHA1-3DES -out "${cert_name}-key.pem"
    openssl req -new -key "${cert_name}-key.pem" -subj "${subject}" -out "${cert_name}.csr"
    openssl x509 -req -in "${cert_name}.csr" -CA root-ca-crt.pem -CAkey root-ca-key.pem -CAcreateserial -sha256 -out "${cert_name}-crt.pem" -days 365

    # Clean up temporary files
    rm "${cert_name}-key-temp.pem" "${cert_name}.csr"
}

function generate_if_not_user_defined {
    # Generate ssl keypair if it wasn't explicitly set by an operator
    local cert_name="${1}"
    local common_name="${2}"
    local key_file="${cert_name}-key.pem"
    local cert_file="${cert_name}-crt.pem"

    if [[ -f "${key_file}" && ! -L "${key_file}" && -f "${cert_file}" && ! -L "${cert_file}" ]]; then
        # If both files exist, we assume they were placed here by the operator
        echo "[INFO] Keeping keypair for ${cert_name}."

    elif [[ ! -e "${key_file}" || -L "${key_file}" || ! -e "${cert_file}" || -L "${cert_file}" ]]; then
        # If the files are symlinks, we assume they were auto-generated by the script.
        # Similarly if they don't exist.
        unlink "${key_file}" &>/dev/null || true
        unlink "${cert_file}" &>/dev/null || true
        generate_certificate "${cert_name}-gen" "${common_name}"
        ln -s "${cert_name}-gen-key.pem" "${key_file}"
        ln -s "${cert_name}-gen-crt.pem" "${cert_file}"

    else
        # Maybe the files got turned into directories by the bind mounts?
        echo "[INFO] Something weird is going on with the files ${cert_name}-{key,crt}.pem"
        echo ">     please investigate manually."
    fi
}

function is_local_setup {
    local hostname="${1}"
    [[ "${hostname}" == "localhost" ]]
}

function link_certbot_certificates {
    # Symlink Certbot certificates
    local cert_name="${1}"
    local hostname="${2}"

    # Spin up again certbot container and change directory permissions to make certificates readable.
    # Symlinks won't work otherwise; 'live' directory has 700 permissions with root ownership.
    docker compose up -d certbot
    eval "docker compose exec -u root certbot sh -c \"
        chmod 0755 /etc/letsencrypt/live /etc/letsencrypt/archive &&
        chmod -R +r /etc/letsencrypt/archive/* /etc/letsencrypt/live/*
    \""

    if [[ $? -ne 0 ]]; then
        echo "[ERROR] Failed to adjust permissions for Let's Encrypt files."
        exit 1
    fi
    echo "[INFO] Permissions adjusted for Let's Encrypt directories."

    echo "[INFO] Linking Certbot certificates for ${hostname} to ${cert_name}."
    local ssl_path="${project_dir}/docker/certbot/conf/live/${hostname}"

    if [[ -d "${ssl_path}" ]]; then
        ln -sf "${ssl_path}/privkey.pem" "${cert_name}-key.pem"
        ln -sf "${ssl_path}/fullchain.pem" "${cert_name}-crt.pem"
    else
        echo "[ERROR] Certbot certificates for ${hostname} not found. Ensure Certbot has generated them."
        exit 1
    fi
}

function append_letsencrypt_ca() {
    local hostname="$1"
    local ca_chain="${project_dir}/docker/certbot/conf/live/${hostname}/chain.pem"
    local root_ca="${project_dir}/ssl/root-ca-crt.pem"

    # Append Let's Encrypt CA to root-ca-crt.pem only if it's not already present.
    # OpenSearch needs it to recognize the new certificates.
    chain_cert_content=$(cat "$ca_chain")
    root_ca_content=$(cat "$root_ca")

    if [[ "$root_ca_content" == *"$chain_cert_content"* ]]; then
        :
    else
        echo -e "\n$chain_cert_content" >> "$root_ca"  # Append correctly formatted certificate.
        echo "[INFO] Appending Let's Encrypt CA to 'root-ca-crt.pem'."
    fi
}

# ------------ #
# Run sequence #
# ------------ #
# Root CA
subject="/C=${country}/ST=${state}/L=${locality}/O=${organization}/OU=${org_unit}/CN=ROOT"
openssl genrsa -out root-ca-key.pem 2048
openssl req -new -x509 -sha256 -key root-ca-key.pem -subj "${subject}" -out root-ca-crt.pem -days 365

# Admin certificate (for securityadmin.sh)
generate_certificate "admin" "admin"

# Node 1 cert (for internal communication)
generate_certificate "node1" "os-node-1"

# Node 2 cert (for internal communication)
generate_certificate "node2" "os-node-2"

# Determine external SSL certificates generation
if is_local_setup "${dashboards_hostname}" || is_local_setup "${opensearch_hostname}"; then
    echo "[INFO] Detected local setup, generating local certificates for Dashboards and Cluster."

    # REST HTTPS cert (for outside communication)
    generate_if_not_user_defined "cluster" "${opensearch_hostname}"

    # Dashboards cert (for outside communication)
    generate_if_not_user_defined "dashboards" "${dashboards_hostname}"

else
    echo "[INFO] Detected live setup, generating Let's Encrypt certificates for Dashboards and Cluster."

    # Select appropriate email argument.
    if [[ -z "${certbot_email}" ]]; then
        email_arg="--register-unsafely-without-email"
    else
        email_arg="--email ${certbot_email}"
    fi

    # Determine certbot staging argument.
    if [[ -z "${certbot_staging}" ]] || [[ ! "${certbot_staging}" =~ ^[01]$ ]]; then
        echo "[WARN] Invalid or unset value for 'CERTBOT_STAGING'. Setting it to 1 (testing mode)."
        certbot_staging="1"
    fi
    if [[ "${certbot_staging}" != "0" ]]; then staging_arg="--staging"; fi

    # Define common certbot command
    certbot_cmd="certbot certonly ${staging_arg:-} ${email_arg} --rsa-key-size ${certbot_rsa_key_size} --standalone --agree-tos --force-renewal"

    ssl_base_path="${project_dir}/docker/certbot/conf/live"

    # Check if Dashboards and Cluster point to the same hostname.
    if [[ "${dashboards_hostname}" == "${opensearch_hostname}" ]]; then
        echo "[INFO] Hostnames for Dashboards and Cluster are the same. Generating a single certificate for both."
        live_crt_path="${ssl_base_path}/${dashboards_hostname}/fullchain.pem"

        if [[ -L "${live_crt_path}" ]]; then
            echo "[WARN] Certificates for ${dashboards_hostname} already exist. Skipping generation."

        else
            eval "docker compose run --rm -p \"80:80\" --entrypoint \"${certbot_cmd} -d ${dashboards_hostname}\" certbot"
            # Symlink Certbot certificates
            link_certbot_certificates "dashboards" "${dashboards_hostname}"
            link_certbot_certificates "cluster" "${opensearch_hostname}"
            append_letsencrypt_ca "${dashboards_hostname}"
        fi

    else
        echo "[INFO] Hostnames for are different. Generating separate certificates for each."
        live_dashboards_crt_path="${ssl_base_path}/${dashboards_hostname}/fullchain.pem"
        live_opensearch_crt_path="${ssl_base_path}/${opensearch_hostname}/fullchain.pem"

        # Attempt to generate certificate for dashboards_hostname.
        if [[ -L "${live_dashboards_crt_path}" ]]; then
            echo "[WARN] Certificates for ${dashboards_hostname} already exist. Skipping generation."

        else
            eval "docker compose run --rm -p \"80:80\" --entrypoint \"${certbot_cmd} -d ${dashboards_hostname}\" certbot"
            link_certbot_certificates "dashboards" "${dashboards_hostname}"
            append_letsencrypt_ca "${dashboards_hostname}"
        fi

        # Attempt to generate certificate for opensearch_hostname.
        if [[ -L "${live_opensearch_crt_path}" ]]; then
            echo "[WARN] Certificates for ${opensearch_hostname} already exist. Skipping generation."

        else
            eval "docker compose run --rm -p \"80:80\" --entrypoint \"${certbot_cmd} -d ${opensearch_hostname}\" certbot"
            link_certbot_certificates "cluster" "${opensearch_hostname}"
            append_letsencrypt_ca "${opensearch_hostname}"
        fi
    fi
fi

# Ensure read permissions to all files in the 'ssl' directory because 'opensearch' user must be able to read them in the container.
# This is especially important for the *-key.pem files (private keys).
chmod -R +r "${project_dir}/ssl"
